(******************************************************************************)
(*   #####   #####    ####      Formalization of the GPS program logic        *)
(*  ##       ##  ##  ##         by Dreyer, Tassarotti, Turon, Vafeiadis       *)
(*  ##  ###  #####    ####                                                    *)
(*  ##   ##  ##          ##     See README.txt for copyright, license, and    *)
(*   #####   ##       ####      installation instructions.                    *)
(******************************************************************************)

Require Import Vbase Relations ClassicalDescription ClassicalChoice.
Require Import List Permutation Sorting extralib. 
Require Import c11 exps.
Require Import GpsFraction GpsSepAlg GpsModel GpsModelLemmas GpsLogic GpsTokens.

Local Open Scope string_scope.

Set Implicit Arguments.

Notation " a |-> b " := (Rpt pe_full a b) (at level 60).

Notation " a ** b " := (Rstar a b) (at level 80, right associativity).

Notation "'ACQ' [ e1 ]" := 
  (Eload RATacq e1) 
  (at level 99, format "'ACQ' [ e1 ]").

Notation "'NA' [ e1 ]" := 
  (Eload RATna e1) 
  (at level 99, format "'NA' [ e1 ]").

Notation "'NA' [ e1 ] := e2" := 
  (Estore WATna e1 e2)
  (at level 99, format "'NA' [ e1 ]  :=  e2").

Notation "'REL' [ e1 ] := e2" := 
  (Estore WATrel e1 e2) 
  (at level 99, format "'REL' [ e1 ]  :=  e2").

Notation "'LET' x <- e 'IN' f" := 
  (Elet e (fun x => f)) 
  (at level 100, f at level 100, format "'[v' 'LET'  x  <-  e  'IN' '/' f ']'").

Notation " e ;; f" := 
  (Elet e (fun _ => f)) 
  (at level 100, f at level 100, format "'[v' e  ;; '/' f ']'").


Definition producer x y :=
  NA[x] := 37 ;;
  REL[y] := 1.

Definition consumer x y :=
  Erepeat ( ACQ[y] ) ;;
  NA[x].


Definition oneshot :=
  LET x <- Ealloc 0  IN
  LET y <- Ealloc 0  IN
  NA[x] := 0 ;;
  REL[y] := 0  ;;
  Efork (producer x y) ;;
  consumer x y.


Parameter decode_pair : nat -> nat * nat.
Parameter encode_pair : nat * nat -> nat.

Axiom enc_dec_pair : forall n, encode_pair (decode_pair n) = n.
Axiom dec_enc_pair : forall n, decode_pair (encode_pair n) = n.

Program Definition MyProt a :=
  {| prot_state := bool ;
     prot_id := a ;
     prot_trans x y := negb x || y |}.
Next Obligation. split; red; ins; destruct x; ins; destruct y; ins. Qed.

Lemma prot_state_of_MyProt ty :  (exists a, ty = MyProt a) -> prot_state ty = bool.
Proof. ins; desf. Qed.

Program Definition PIE := 
  {| interp_prot := fun ty s v => 
                      match excluded_middle_informative (exists a, ty = MyProt a) return _ with
                        | left EQ =>
                            if (eq_rect _ id s _ (prot_state_of_MyProt EQ)) then 
                              Rescr (mkEscrowTy (prot_id ty)) 
                            else 
                              Rpure (v = 0)
                        | _ => Rfalse
                      end
   ; interp_exch := fun sigma => 
                      match sigma return _ with
                        | mkEscrowTy a =>
                          (Tok (fst (decode_pair a)), Rpt pe_full (snd (decode_pair a)) 37)
                      end
  |}.
Next Obligation. by desf; intro M; apply star_tok_tok in M. Qed.
Next Obligation. by desf; intro M; apply star_pt_pt in M. Qed.


Lemma producer_spec x y (g : nat) : 
  triple PIE 
         (x |-> 0 ** Rat y (MyProt (encode_pair (g,x))) false) 
         (producer x y) 
         (fun _ => Rtrue).
Proof.
  eapply rule_let. 
    by eapply rule_frame, rule_write_na.
  simpl; intros _.
  eapply rule_pconseq; ins.
    by eapply gimp_frame, gimp_eintro2 with (sigma := mkEscrowTy (encode_pair (g,x))); 
       ins; rewrite dec_enc_pair; ins.
  eapply rule_conseq; ins; 
    [by apply gimp_refl| |by eapply gimp_implies, implies_true].
  rewrite RstarC; apply rule_store_rel with (s'' := true) (Q := Rtrue); try done;
    [|by ins; rewrite orbT].
  rewrite Rstar_true.
  ins; desf; desf; rewrite <- ?eq_rect_eq in *; desf; try (by destruct n; vauto).
  by eapply gimp_refl.
Qed. 

Lemma consumer_spec x y (g : nat) : 
  triple PIE 
         (Rat y (MyProt (encode_pair (g, x))) false ** Tok g)
         (consumer x y) 
         (fun r => Rpure (r = 37)).
Proof.
  eapply rule_let.
    eapply rule_repeat.
    eapply rule_frame.
    rewrite Rat_dupl.
    eapply rule_read_acq with (QQ := fun _ v => if v == 0 then Rtrue
                                              else Rescr (mkEscrowTy (encode_pair (g, x)))); 
      try done.
    red; intros; rewrite RstarC, Rstar_at in SAT.
    ins; desf; desf; try (by destruct n; vauto); rewrite <- eq_rect_eq in *; ins; desf;
    rewrite ?res_escr_strip; eauto using res_get_strip_at; try (by destruct r).
    ins; apply Rstar_mono; ins; apply RexistsE; ins.
    by apply Rstar_forget_l, Rstar_forget_r.
  ins; apply triple_conj_pure; ins; desf; desf.
  rewrite RstarC, star_exists; apply rule_exists; ins. 

  eapply rule_conseq; try rewrite <- app_nil_end; ins.
      rewrite RstarC, !RstarA; do 2 eapply gimp_frame_l; rewrite RstarC, Rbox_escr.
      by eapply gimp_eelim; ins; rewrite dec_enc_pair; ins.
    by rewrite RstarC, RstarA, RstarAC; eapply rule_frame, rule_read_na.
  by ins; eapply gimp_implies; red; ins; desf.
Qed.

Lemma oneshot_spec : 
  triple PIE Rtrue oneshot (fun r => Rpure (r = 37)).
Proof.
  eapply rule_let; [by apply rule_alloc_one|]; intro x.
  eapply rule_let; [|intro y].
    rewrite <- (Rstar_true (Runinit x)), RstarC.
    by eapply rule_frame, rule_alloc_one.
  apply new_tok; intro g.
  eapply rule_let.
    rewrite RstarA, RstarAC; eapply rule_frame.
    by eapply rule_write_na_uninit.
  simpl; intros _. 
  eapply rule_let.
    rewrite RstarAC.
    eapply rule_frame.
    eapply rule_conseq; ins; [| |by apply gimp_refl].
    2: by eapply rule_store_rel_uninit with (tau := MyProt (encode_pair (g,x))).
    instantiate (1 := false).
    eapply gimp_implies; ins; desf; desf; try (by destruct n; vauto);
    rewrite <- eq_rect_eq in *; ins.
    by split; ins; try intro; desf; repeat eexists; eauto using res_plus_emp_r; ins.
  simpl; intros _.
  eapply rule_let.
    rewrite RstarAC, Rat_dupl, RstarA, <- RstarA.
    by eapply rule_frame, rule_fork, producer_spec.
  simpl; intros _.
  rewrite RstarC, Rstar_true; apply consumer_spec.
Qed.
